import re
import datetime
from pymongo import MongoClient
from bson import ObjectId
from .exception import RecordException, StructureException

__all__ = ('get_database', 'Record', 'Structure')


def get_database(db_name, host, port=27017):
    return MongoClient(host, port)[db_name]


class Structure(dict):
    __store = {}

    def __init__(self, *args, **kwargs):
        super(Structure, self).__init__(*args, **kwargs)
        self.__dict__ = self

    def init_from_dict(self, data):
        if not isinstance(data, dict):
            data = {}

        store = {}
        for key, value in self.__dict__.items():
            if '__store' in key:
                continue

            result = data.get(key)
            if not result:
                result = value
            store[key] = result

        self.__store = store
        self._validate()

    def _validate(self):
        pass

    def assign_to_store(self, key, value):
        if key in self.__store.keys() or key in "_id":
            self.__store[key] = value
            return True
        raise StructureException("value is invalid type, key : {0}".format(value))

    def get_from_store(self, key):
        if key in self.__store.keys():
            return self.__store[key]
        raise StructureException("{0} is not a registered".format(key))

    def to_dict(self):
        return self.__store

    def to_mongo(self):
        store = self.to_dict()

        now = datetime.datetime.now()
        if not 'created_at' in store.keys():
            store['created_at'] = now
        store['modified_at'] = now

        if '_id' in store.keys():
            del store['_id']
        return store


class Record:
    struct = None

    class Meta:
        database = None

    def __init__(self, key, data=None):
        self._key = key
        self._init_from_dict(data)

    def _init_from_dict(self, data):
        if not isinstance(self.struct, Structure):
            raise RecordException("{0} struct is not a defined".format(self.__class__.__name__))

        # initialize store data
        self.struct.init_from_dict(data)

    def key(self):
        return self._key

    def __str__(self):
        return self.__name__

    def __getattr__(self, key):
        if key in self.struct.keys():
            return self.struct.get_from_store(key)
        else:
            return super(Record, self).__getattr__(key)

    def __setattr__(self, key, value):
        if key in self.struct.keys():
            self.struct.assign_to_store(key, value)
        else:
            super(Record, self).__setattr__(key, value)

    @classmethod
    def colname(cls):
        return re.sub('(?!^)([A-Z]+)', r'_\1', cls.__name__).lower().__str__()

    @classmethod
    def collection(cls):
        return cls.Meta.database[cls.colname()]

    @classmethod
    def new(cls, data=None):
        return cls(None, data)

    @classmethod
    def get(cls, key, *args, **kwargs):
        data = cls.collection().find_one({'_id': ObjectId(key)}, *args, **kwargs)
        if not data:
            return None
        return cls(key, data)

    @classmethod
    def find(cls, *args, **kwargs):
        results = cls.collection().find(*args, **kwargs)

        insances = []
        for _, v in enumerate(results):
            insances += [cls(v['_id'].__str__(), v)]
        return insances

    def save(self):
        if not self.key():
            return self.insert()
        return self.update()

    def insert(self):
        result = self.collection().insert_one(self.struct.to_mongo())

        self._key = result.inserted_id.__str__()
        self.struct.assign_to_store('_id', self.key())
        return True

    def update(self, upsert=False):
        if not self.key():
            return self.insert()

        self.collection().update_one({'_id': self.key()}, {'$set': self.struct.to_mongo()}, upsert=upsert)
        return True

    def delete(self):
        if not self.key():
            return False

        self.collection().delete_one({'_id': ObjectId(self.key())})
        return True



import unittest
from nose.tools import eq_

import icePick
from tests.config import ORDER_URL, ORDER_UA, PARSE_HTML


class TestOrderModel(icePick.Order):
    pass


class TestOrder(unittest.TestCase):
    def setUp(self):
        self.order = TestOrderModel(ORDER_URL, ORDER_UA, method=TestOrderModel.Method.POST)

    def test_validate(self):
        eq_('utf-8', self.order.charset)
        eq_('text/html', self.order.content_type)
        eq_(icePick.Order.Method.POST, self.order.method)

        new_order = TestOrderModel(ORDER_URL, ORDER_UA, method="hogefuga", charset='uTF-8', content_type="TEXT/html")
        eq_('utf-8', new_order.charset)
        eq_('text/html', new_order.content_type)
        eq_(icePick.Order.Method.GET, new_order.method)

    def test_user_agent(self):
        eq_(ORDER_UA, self.order.user_agent)

    def test_get_headers(self):
        headers = {
            'User-agent': ORDER_UA,
            'Content-Type': "text/html; utf-8"
        }
        eq_(headers, self.order.get_headers())

    def test_parse(self):
        eq_({}, self.order.parse(PARSE_HTML))

    def test_save(self):
        eq_(None, self.order.save({"foo": "bar"}))



import unittest
from nose.tools import ok_, eq_

import datetime
from pymongo import MongoClient

from tests.config import DB_HOST, DB_PORT, DB_NAME
from icePick.record import get_database, Structure, Record


db = get_database(DB_NAME, DB_HOST, DB_PORT)


class TestStructureModel(Structure):
    pass


class TestRecordModel(Record):
    struct = TestStructureModel(
        string=str(),
        intger=int(),
        boolean=bool(),
        list_=list(),
        dictionary=dict(),
        dt=datetime.datetime.now()
    )

    class Meta:
        database = db


class TestRecord(unittest.TestCase):
    def setUp(self):
        self.record = TestRecordModel.new()

    def tearDown(self):
        m = MongoClient(DB_HOST, DB_PORT)
        m.drop_database(DB_NAME)

    def test_colname(self):
        eq_('test_record_model', self.record.colname())

    def test_attrs(self):
        new_str = "test_setattr"
        self.record.string = new_str

        eq_(new_str, self.record.string)

    def test_new(self):
        eq_(None, self.record.key())

        new_record = TestRecordModel.new({
            "string": "new_string"
        })
        eq_("new_string", new_record.string)

    def test_insert(self):
        eq_(None, self.record.key())

        self.record.string = "new_str"
        self.record.insert()

        eq_("new_str", self.record.string)
        self.assertNotEqual(None, self.record.key())

    def test_update(self):
        self.record.insert()
        self.assertNotEqual(None, self.record.key())

        self.record.string = "new_str"
        self.record.update()
        eq_("new_str", self.record.string)

    def test_save(self):
        self.record.string = "new_str"
        self.record.save()

        self.assertNotEqual(None, self.record.key())
        eq_("new_str", self.record.string)

        self.record.string = "update"
        self.record.save()
        eq_("update", self.record.string)

    def test_get(self):
        self.record.string = "new_str"
        self.record.save()

        exist_record = TestRecordModel.get(self.record.key())
        eq_(exist_record.key(), self.record.key())
        eq_(exist_record.string, self.record.string)

    def test_find(self):
        result = TestRecordModel.find()
        eq_(0, result.__len__())
        self.record.save()

        result = TestRecordModel.find()
        eq_(1, result.__len__())
        eq_(result[0].key(), self.record.key())

    def test_delete(self):
        self.record.save()

        result = TestRecordModel.find()
        eq_(1, result.__len__())

        self.record.delete()

        result = TestRecordModel.find()
        eq_(0, result.__len__())



DB_HOST = '127.0.0.1'
DB_PORT = 27017
DB_NAME = 'test_database'

PARSE_HTML = """<!DOCTYPE html>
<html lang="en">
    <head>
        <meta charset='utf-8'/>
        <title>TestHTML</title>
    </head>
    <body>
        <div class="container">
            <p>HTML parse test</p>
        </div>
    </body>
</html>"""


ORDER_HOST = "127.0.0.1"
ORDER_PORT = "8888"
ORDER_URL = "http://{0}:{1}".format(ORDER_HOST, ORDER_PORT)
ORDER_UA = "Mozilla/5.0 (Windows NT 6.3; Trident/7.0; rv:11.0) like Gecko"



import unittest
from nose.tools import eq_
from pymongo import MongoClient
from threading import Thread
from http.server import HTTPServer, SimpleHTTPRequestHandler

import icePick
from tests.config import DB_HOST, DB_PORT, DB_NAME, ORDER_HOST, ORDER_PORT, ORDER_URL, ORDER_UA, PARSE_HTML
from tests.test_parser import TestParserModel as _TestParserModel

db = icePick.get_database(DB_NAME, DB_HOST, DB_PORT)


class TestRequestHandler(SimpleHTTPRequestHandler):
    def do_GET(self):
        body = str.encode(PARSE_HTML)
        self.send_response(200)
        self.send_header('Content-type', 'text/html; charset=utf-8')
        self.send_header('Content-length', len(body))
        self.end_headers()
        self.wfile.write(body)


class TestParserModel(_TestParserModel):
    pass


class TestRecordModel(icePick.Record):
    struct = icePick.Structure(title=str(), charset=str(), text=str())

    class Meta:
        database = db


class TestOrderModel(icePick.Order):
    def parse(self, html):
        parser = TestParserModel(html)
        return parser.run()

    def save(self, result):
        record = TestRecordModel.new(result)
        return record.save()


class TestPickerModel(unittest.TestCase):
    def setUp(self):
        order = TestOrderModel(ORDER_URL, ORDER_UA)
        self.picker = icePick.Picker([order])

        self.httpd = HTTPServer((ORDER_HOST, int(ORDER_PORT)), TestRequestHandler)

        thread = Thread(target=self.httpd.serve_forever)
        thread.start()

    def tearDown(self):
        thread = Thread(target=self.httpd.shutdown)
        thread.start()

        m = MongoClient(DB_HOST, DB_PORT)
        m.drop_database(DB_NAME)

    def test_run(self):
        result = TestRecordModel.find()
        eq_(0, result.__len__())

        self.picker.run()

        result = TestRecordModel.find()
        eq_(1, result.__len__())

        record = result[0]
        eq_('TestHTML', record.title)
        eq_('utf-8', record.charset)
        eq_('HTML parse test', record.text)



import unittest
from nose.tools import eq_

import icePick
from tests.config import PARSE_HTML


class TestParserModel(icePick.Parser):
    def serialize(self):
        title = self.bs.find('title').text
        charset = self.bs.find('meta').get('charset')
        text = self.bs.find('p').text

        return {
            "title": title,
            "charset": charset,
            "text": text,
        }


class TestParser(unittest.TestCase):
    def setUp(self):
        self.parser = TestParserModel(PARSE_HTML)

    def test_lang(self):
        eq_('en', self.parser.lang)

    def test_serialize(self):
        result = self.parser.serialize()

        eq_('TestHTML', result['title'])
        eq_('utf-8', result['charset'])
        eq_('HTML parse test', result['text'])

    def test_run(self):
        result = self.parser.run()

        eq_('TestHTML', result['title'])
        eq_('utf-8', result['charset'])
        eq_('HTML parse test', result['text'])



from bs4 import BeautifulSoup

__all__ = ('Parser')


class Parser:
    def __init__(self, html):
        self.bs = BeautifulSoup(html)

    @property
    def lang(self):
        return self.bs.find('html').get('lang')

    @property
    def can_allow_parse(self):
        return True

    def run(self):
        if not self.can_allow_parse:
            return {}
        return self.serialize()

    def serialize(self):
        return {}



import asyncio
import aiohttp

__all__ = ('Picker')


class Picker:
    def __init__(self, orders):
        self.orders = orders
        self.loop = asyncio.get_event_loop()

    def run(self, is_event_close=True):
        result = asyncio.wait(self.get_orders_with_async(), loop=self.loop)
        self.loop.run_until_complete(result)

        if is_event_close:
            self.loop.close()

    def get_orders_with_async(self):
        return [asyncio.async(self._dispatch(order)) for order in self.orders]

    @asyncio.coroutine
    def _dispatch(self, order):
        html = yield from self._download(url=order.url, method=order.method.value, headers=order.get_headers())
        result = order.parse(html.decode(order.charset, 'ignore'))
        order.save(result)

    @asyncio.coroutine
    def _download(self, method, *args, **kwargs):
        response = yield from aiohttp.request(method, *args, **kwargs)
        return (yield from response.read_and_close())



#!/usr/bin/env python
"""
IcePick
===================

IcePick is a All in one Package library for easy Scraping

--------------

Concept
-------

-  Lightweight Scraping Library
-  All in one Package library for easy Scraping

Requirements
------------

-  Python 3.4 or later(not support 2.x)
-  MongoDB

Dependencies Libraries
----------------------

-  aiohttp
-  beautifulsoup4
-  pymongo >= 3.0
-  nose

Usage
-----

Scraping Flow,

::

    Your Scraping Order(Order) -> Do Scraping(Picker) -> HTML Parse(Parser) -> Save in Database(Record)

Example
-------

get a my repository filenames

.. code:: python


    import icePick

    db = icePick.get_database('icePick_example', 'localhost')


    class GithubRepoParser(icePick.Parser):
        def serialize(self):
            result = {
                "files": [],
            }

            for v in self.bs.find_all(class_="js-directory-link"):
                result['files'] += [v.text]
            return result


    class GithubRepoRecord(icePick.Record):
        struct = icePick.Structure(files=list())

        class Meta:
            database = db


    class GithubRepoOrder(icePick.Order):
        def parse(self, html):
            parser = GithubRepoParser(html)
            return parser.run()

        def save(self, result):
            record = GithubRepoRecord.new(result)
            return record.save()


    def main():
        document = {
            'url': 'https://github.com/teitei-tk/ice-pick/tree/master',
            'ua': 'Mozilla/5.0 (Windows NT 6.3; Trident/7.0; rv:11.0) like Gecko',
        }

        print('---download start---')
        order = GithubRepoOrder(document.get('url'), document.get('ua'))
        picker = icePick.Picker([order])
        picker.run()
        print("---finish---")

    if __name__ == "__main__":
        main()

::

    >>> import icePick
    >>> db = icePick.get_database('icePick_example', 'localhost')
    >>> class GithubRepoRecord(icePick.Record):
    ...     struct = icePick.Structure(files=list())
    ...     class Meta:
    ...         database = db
    ...
    >>> records = GithubRepoRecord.find()
    >>> records[0].files
    ['example', 'icePick', 'tests', 'LICENSE', 'README.md', 'circle.yml', 'requirements.txt']
    >>>

TODO
----

-  Crawling
-  Document

LICENSE
-------

-  MIT

"""

try:
    import setuptools
    from setuptools import setup, find_packages
except ImportError:
    import sys
    print("Please install setuptools.")
    sys.exit(1)


classifiers = [
    'Development Status :: 3 - Alpha',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3 :: Only',
    'Topic :: Software Development',
    'Topic :: Software Development :: Libraries',
    'Topic :: Software Development :: Libraries :: Python Modules',
]

setup(
    name='icePick',
    version='0.0.1',
    description='icePick is a All in one Package library for easy Scraping',
    long_description=__doc__,
    author='teitei-tk',
    author_email='teitei.tk@gmail.com',
    url='https://github.com/teitei-tk/ice-pick',
    packages=find_packages(),
    include_package_data=True,
    license='MIT',
    classifiers=classifiers,
    install_requires=open('requirements.txt').read().splitlines(),
    keywords=['scraping'],
    download_url='https://github.com/teitei-tk/ice-pick/archive/master.tar.gz'
)



import enum
from .parser import Parser

__all__ = ('Order')


class Order:
    class Method(enum.Enum):
        GET = "GET"
        POST = "POST"
        PUT = "POST"
        DELETE = "POST"

    def __init__(self, url, ua, method=None, charset='utf-8', content_type='text/html'):
        self.url = url
        self.ua = ua
        self.method = method
        self.charset = charset
        self.content_type = content_type

        self._validate()

    def _validate(self):
        self.charset = self.charset.lower()
        self.content_type = self.content_type.lower()

        if not isinstance(self.method, self.Method):
            self.method = self.Method.GET

    @property
    def user_agent(self):
        return self.ua

    def get_headers(self):
        headers = {
            'User-agent': self.user_agent,
            'Content-Type': "{0}; {1}".format(self.content_type, self.charset)
        }
        return headers

    def parse(self, html):
        """
        parser = Parser(html)
        return parser.run()
        """
        return {}

    def save(self, result):
        """
        record = Record.new()
        record.save(result)
        """
        pass



import sys
import os
sys.path.append(os.path.dirname(os.path.abspath(__file__)) + '/../')

import icePick

db = icePick.get_database('icePick_example', 'localhost')


class GithubRepoParser(icePick.Parser):
    def serialize(self):
        result = {
            "files": [],
        }

        for v in self.bs.find_all(class_="js-directory-link"):
            result['files'] += [v.text]
        return result


class GithubRepoRecord(icePick.Record):
    struct = icePick.Structure(files=list())

    class Meta:
        database = db


class GithubRepoOrder(icePick.Order):
    def parse(self, html):
        parser = GithubRepoParser(html)
        return parser.run()

    def save(self, result):
        record = GithubRepoRecord.new(result)
        return record.save()


def main():
    document = {
        "url": "https://github.com/teitei-tk/ice-pick/tree/master",
        "ua": "Mozilla/5.0 (Windows NT 6.3; Trident/7.0; rv:11.0) like Gecko",
    }

    print('---donload start---')
    order = GithubRepoOrder(document.get('url'), document.get('ua'))
    picker = icePick.Picker([order])
    picker.run()
    print("---finish---")

if __name__ == "__main__":
    main()
